/*
* ============================================================================
*  Name        : loopbackaudiocontrol.cpp
*  Part of     : mccexample / mccexengine
*  Description : CLoopbackAudioControl class implementation
*  Version     : %version: %
*
*  Copyright  2008-2009 Nokia.  All rights reserved.
*  This material, including documentation and any related computer
*  programs, is protected by copyright controlled by Nokia.  All
*  rights are reserved.  Copying, including reproducing, storing,
*  adapting or translating, any or all of this material requires the
*  prior written consent of Nokia.  This material also contains
*  confidential information which may not be disclosed to others
*  without the prior written consent of Nokia.
* ============================================================================
* Template version: 4.1.1
*/

#include <mmccinterface.h>
#include <mmcccodecinformation.h>

#include "loopbackaudiocontrol.h"
#include "mccnetworksession.h"
#include "mcceventhandler.h"
#include "uplinkaudiostream.h"
#include "downlinkaudiostream.h"
#include "streamstatemachine.h"
#include "tcodecsettings.h"
#include "mccexpanic.pan"

// ======== MEMBER FUNCTIONS ========

// ---------------------------------------------------------------------------
// CLoopbackAudioControl::CLoopbackAudioControl
// ---------------------------------------------------------------------------
//
CLoopbackAudioControl::CLoopbackAudioControl( CMccInterface& aMcc,
    CMccEventHandler& aEventHandler,
    const TNetworkSettings& aSettings ) : 
    iMcc( aMcc ), 
    iEventHandler( aEventHandler ),
    iNwSettings( aSettings )
    {
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::~CLoopbackAudioControl
// ---------------------------------------------------------------------------
//
CLoopbackAudioControl::~CLoopbackAudioControl()
    {
    iEventHandler.RemoveSession( iSession );
    iAudioStreams.ResetAndDestroy();
    delete iSession;
    delete iStateMachine;
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::NewLC
// ---------------------------------------------------------------------------
//
CLoopbackAudioControl* CLoopbackAudioControl::NewLC( CMccInterface& aMcc,
    CMccEventHandler& aEventHandler,
    MMccExEngineObserver& aObserver,
    const TNetworkSettings& aSettings,
    const TCodecSettings& aCodec )
    {
    CLoopbackAudioControl* self = 
        new ( ELeave ) CLoopbackAudioControl( aMcc,
            aEventHandler, aSettings );
    
    CleanupStack::PushL( self );
    self->ConstructL( aObserver, aCodec );
    return self;
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::NewL
// ---------------------------------------------------------------------------
//
CLoopbackAudioControl* CLoopbackAudioControl::NewL( CMccInterface& aMcc,
    CMccEventHandler& aEventHandler,
    MMccExEngineObserver& aObserver,
    const TNetworkSettings& aSettings,
    const TCodecSettings& aCodec )
    {
    CLoopbackAudioControl* self = 
        CLoopbackAudioControl::NewLC( aMcc, aEventHandler,
            aObserver, aSettings, aCodec );
    
    CleanupStack::Pop( self );
    return self;
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::ConstructL
// ---------------------------------------------------------------------------
//
void CLoopbackAudioControl::ConstructL( MMccExEngineObserver& aObserver,
    const TCodecSettings& aCodec )
    {
    RPointerArray<CMccCodecInformation> codecs;
    CleanupResetAndDestroyPushL( codecs );
    User::LeaveIfError( iMcc.GetCapabilities( codecs ) );
    
    // Find the used codec and set the values given from settings.
    CMccCodecInformation* codec = FindCodecByNameL( aCodec.iCodecName,
        codecs );
    
    User::LeaveIfError( codec->SetPTime( aCodec.iPTime ) );
    User::LeaveIfError( codec->SetMaxPTime( aCodec.iMaxPTime ) );
    User::LeaveIfError(
        codec->SetJitterBufBufferLength( aCodec.iJitterBufferSize ) );
    User::LeaveIfError( codec->EnableVAD( aCodec.iVadCn ) );
    
    // Now that settings are OK - start creating the rest of the resources.
    iStateMachine = CStreamStateMachine::NewL();
    iSession = CMccNetworkSession::NewL( iMcc, aObserver, 
        iNwSettings.iMediaQoS, iNwSettings.iRtpPort );
    
    iEventHandler.AddSessionL( iSession );
    
    // The linkID at this point is actually correct, only thing is that the
    // link is not up until MCC sends 'Link created' - event. Thus we can
    // safely start creating the streams at this point. Also, the streams do 
    // not reserve any MCC resources at this point. That must be done 
    // separately and later once the link creation is finished.
    CAudioStreamBase* stream( NULL );
    
    stream = CUplinkAudioStream::NewLC( *iStateMachine,
            *codec, iMcc, iSession->SessionId(), iSession->LinkId() );
    stream->SetRtcpUsage( iNwSettings.iRtcpEnabled );
    
    iAudioStreams.AppendL( stream );
    iSession->AddStreamL( *stream );
    CleanupStack::Pop( stream );
    stream = NULL;
    
    stream = CDownlinkAudioStream::NewLC( *iStateMachine,
            *codec, iMcc, iSession->SessionId(), iSession->LinkId() );
    stream->SetRtcpUsage( iNwSettings.iRtcpEnabled );
    
    iAudioStreams.AppendL( stream );
    iSession->AddStreamL( *stream );
    CleanupStack::Pop( stream );
    stream = NULL;
    
    CleanupStack::PopAndDestroy( &codecs );
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::PrepareAudioL
// ---------------------------------------------------------------------------
//
void CLoopbackAudioControl::PrepareAudioL()
    {
    TInetAddr address( iNwSettings.iRemoteAddress );
    address.SetPort( iNwSettings.iRtpPort );
    
    iSession->SetRemoteAddressL( address );
    
    TInt count( iAudioStreams.Count() );
    while ( count )
        {
        count -= 1;
        iAudioStreams[count]->PrepareStreamL();
        }
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::StartAudioL
// ---------------------------------------------------------------------------
//
void CLoopbackAudioControl::StartAudioL()
    {
    TInt count( iAudioStreams.Count() );
    while ( count )
        {
        count -= 1;
        iAudioStreams[count]->StartStreamL();
        }
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::PauseAudioL
// ---------------------------------------------------------------------------
//
void CLoopbackAudioControl::PauseAudioL()
    {
    const TInt count( iAudioStreams.Count() );
    TInt k = 0;
    
    // First stop the uplink streams so the downlink goes into standby state
    // and after that, stop it explicitly
    for ( k = 0; k < count; k++ )
        {
        if ( KMccAudioUplinkStream == iAudioStreams[k]->StreamType() )
            {
            iAudioStreams[k]->PauseStreamL();
            }
        }
    
    for ( k = 0; k < count; k++ )
        {
        if ( KMccAudioDownlinkStream == iAudioStreams[k]->StreamType() )
            {
            iAudioStreams[k]->PauseStreamL();
            }
        }
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::ResumeAudioL
// ---------------------------------------------------------------------------
//
void CLoopbackAudioControl::ResumeAudioL()
    {
    TInt count( iAudioStreams.Count() );
    while ( count )
        {
        count -= 1;
        iAudioStreams[count]->ResumeStreamL();
        }
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::StopAudioL
// ---------------------------------------------------------------------------
//
void CLoopbackAudioControl::StopAudioL()
    {
    const TInt count( iAudioStreams.Count() );
    TInt k = 0;
    
    // First stop the uplink streams so the downlink goes into standby state
    // and after that, stop it explicitly
    for ( k = 0; k < count; k++ )
        {
        if ( KMccAudioUplinkStream == iAudioStreams[k]->StreamType() )
            {
            iAudioStreams[k]->StopStreamL();
            }
        }
    
    for ( k = 0; k < count; k++ )
        {
        if ( KMccAudioDownlinkStream == iAudioStreams[k]->StreamType() )
            {
            iAudioStreams[k]->StopStreamL();
            }
        }
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::SetVolumeL
// ---------------------------------------------------------------------------
//
void CLoopbackAudioControl::SetVolumeL( TInt aVolume )
    {
    // Volume is applied to all downlink audio streams.
    TInt idx = iAudioStreams.Count();
    while( idx )
        {
        idx -= 1;
        if ( KMccAudioDownlinkStream == iAudioStreams[idx]->StreamType() )
            {
            static_cast<CDownlinkAudioStream*>(
                    iAudioStreams[idx] )->SetVolumeL( aVolume );
            }
        }
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::VolumeL
// ---------------------------------------------------------------------------
//
TInt CLoopbackAudioControl::VolumeL()
    {
    TInt idx = iAudioStreams.Count();
    TInt volume( KMaxTInt );
    while( idx )
        {
        idx -= 1;
        if ( KMccAudioDownlinkStream == iAudioStreams[idx]->StreamType() )
            {
            volume = static_cast<CDownlinkAudioStream*>(
                    iAudioStreams[idx] )->VolumeL();
            break;
            }
        }
    
    return volume;
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::MaxVolumeL
// ---------------------------------------------------------------------------
//
TInt CLoopbackAudioControl::MaxVolumeL()
    {
    TInt idx = iAudioStreams.Count();
    TInt maxvolume( KMaxTInt );
    while( idx )
        {
        idx -= 1;
        if ( KMccAudioDownlinkStream == iAudioStreams[idx]->StreamType() )
            {
            maxvolume = static_cast<CDownlinkAudioStream*>(
                    iAudioStreams[idx] )->MaxVolumeL();
            break;
            }
        }
    
    return maxvolume;
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::SetMuteStateL
// ---------------------------------------------------------------------------
//
void CLoopbackAudioControl::SetMuteStateL( TBool aMuteState )
    {
    TInt idx = iAudioStreams.Count();
    while( idx )
        {
        idx -= 1;
        if ( KMccAudioUplinkStream == iAudioStreams[idx]->StreamType() )
            {
            static_cast<CUplinkAudioStream*>(
                    iAudioStreams[idx] )->SetMuteStateL( aMuteState );
            }
        }
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::SetAudioRoutingL
// ---------------------------------------------------------------------------
//
void CLoopbackAudioControl::SetAudioRoutingL( TUint32 aRouting )
    {
    // Audio routing is applied to all downlink audio streams.
    TInt idx = iAudioStreams.Count();
    while( idx )
        {
        idx -= 1;
        if ( KMccAudioDownlinkStream == iAudioStreams[idx]->StreamType() )
            {
            static_cast<CDownlinkAudioStream*>(
                    iAudioStreams[idx] )->SetCurrentAudioRouteL( aRouting );
            }
        }
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::AudioRoutingL
// ---------------------------------------------------------------------------
//
TUint32 CLoopbackAudioControl::AudioRoutingL()
    {
    // Audio routing is only applicable to downlink audio streams.
    TInt idx = iAudioStreams.Count();
    TUint32 routing( KMaxTUint32 );
    while( idx )
        {
        idx -= 1;
        if ( KMccAudioDownlinkStream == iAudioStreams[idx]->StreamType() )
            {
            routing = static_cast<CDownlinkAudioStream*>(
                    iAudioStreams[idx] )->CurrentAudioRouteL();
            break;
            }
        }
    
    return routing;
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::State
// ---------------------------------------------------------------------------
//
TControlState CLoopbackAudioControl::State()
    {
    const TInt count( iAudioStreams.Count() );
    if ( count )
        {
        TControlState state ( iAudioStreams[0]->State() );
        for ( TInt k = 1; k < count; k++ )
            {
            if ( state != iAudioStreams[k]->State() )
                {
                return EStateChanging;
                }
            }
        
        return state;
        }
    else
        {
        MccExPanic( EMccExUnknownState );
        return EStateChanging;
        }
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::ReleaseAudioControl
// ---------------------------------------------------------------------------
//
void CLoopbackAudioControl::ReleaseAudioControl()
    {
    delete this;
    }


// ---------------------------------------------------------------------------
// CLoopbackAudioControl::FindCodecByNameL
// ---------------------------------------------------------------------------
//
CMccCodecInformation* CLoopbackAudioControl::FindCodecByNameL(
    const TDesC8& aName,
    const RPointerArray<CMccCodecInformation>& aCapabilities )
    {
    CMccCodecInformation* codec( NULL );
    TInt codecs( aCapabilities.Count() );
    while( codecs && !codec )
        {
        codecs -= 1;
        if ( 0 == aCapabilities[codecs]->SdpName().Compare( aName ) )
            {
            codec = aCapabilities[codecs];
            }
        }
    
    // Precondition check.
    __ASSERT_ALWAYS( codec, User::Leave( KErrNotFound ) );
    return codec;
    }
