/*
* ============================================================================
*  Name        : mccnetworksession.cpp
*  Part of     : mccexample / mccexengine
*  Description : MCC network session implementation
*  Version     : %version: %
*
*  Copyright  2008-2009 Nokia.  All rights reserved.
*  This material, including documentation and any related computer
*  programs, is protected by copyright controlled by Nokia.  All
*  rights are reserved.  Copying, including reproducing, storing,
*  adapting or translating, any or all of this material requires the
*  prior written consent of Nokia.  This material also contains
*  confidential information which may not be disclosed to others
*  without the prior written consent of Nokia.
* ============================================================================
* Template version: 4.1.1
*/

#include <mmccinterface.h>
#include <mmccevents.h>

#include "mmccexengineobserver.h"
#include "mccnetworksession.h"
#include "streambase.h"
#include "mccexpanic.pan"

// LOCAL CONSTANTS
const TInt KExDefaultIap = -1;

// ======== MEMBER FUNCTIONS ========

// ---------------------------------------------------------------------------
// CMccNetworkSession::CMccNetworkSession
// ---------------------------------------------------------------------------
//
CMccNetworkSession::CMccNetworkSession( CMccInterface& aMcc,
    MMccExEngineObserver& aObserver, TInt aMediaQos, TInt aPort  ) : 
    iMcc( aMcc ),
    iObserver( aObserver ),
    iSessionId( KMaxTUint32 ),
    iLinkId( KMaxTUint32 ),
    iMediaQos( aMediaQos ),
    iLocalPort( aPort )
    {
    }


// ---------------------------------------------------------------------------
// CMccNetworkSession::~CMccNetworkSession
// ---------------------------------------------------------------------------
//
CMccNetworkSession::~CMccNetworkSession()
    {
    // Note the following:
    // For demonstration purposes we first close the link, then the session.
    // For a little bit better performance we could call only "CloseSession()"
    // or even just delete iInterface like CMccExEngine does.
    
    if ( KMaxTUint32 != iLinkId && KMaxTUint32 != iSessionId && iLinkUp )
        {
        iMcc.CloseLink( iSessionId, iLinkId );
        }
    
    if ( KMaxTUint32 != iSessionId )
        {
        iMcc.CloseSession( iSessionId );
        }
    
    // Only close the array as we do not own the instances.
    iStreams.Close();
    }


// ---------------------------------------------------------------------------
// CMccNetworkSession::NewLC
// ---------------------------------------------------------------------------
//
CMccNetworkSession* CMccNetworkSession::NewLC( CMccInterface& aMcc,
    MMccExEngineObserver& aObserver, TInt aMediaQos, TInt aPort )
    {
    CMccNetworkSession* self =
        new ( ELeave ) CMccNetworkSession(
            aMcc, aObserver, aMediaQos, aPort );
    
    CleanupStack::PushL( self );
    self->ConstructL();
    return self;
    }


// ---------------------------------------------------------------------------
// CMccNetworkSession::NewL
// ---------------------------------------------------------------------------
//
CMccNetworkSession* CMccNetworkSession::NewL( CMccInterface& aMcc,
    MMccExEngineObserver& aObserver, TInt aMediaQos, TInt aPort )
    {
    CMccNetworkSession* self = CMccNetworkSession::NewLC(
        aMcc, aObserver, aMediaQos, aPort );
    
    CleanupStack::Pop( self );
    return self;
    }


// ---------------------------------------------------------------------------
// CMccNetworkSession::ConstructL
// ---------------------------------------------------------------------------
//
void CMccNetworkSession::ConstructL()
    {
    User::LeaveIfError( iMcc.CreateSession( iSessionId ) );
    
    TMccNetSettings netSettings;
    netSettings.iLocalAddress.SetPort( iLocalPort );
    netSettings.iIapId = KExDefaultIap;
    netSettings.iMediaQosValue = iMediaQos;
    
    User::LeaveIfError( iMcc.CreateLink( 
            iSessionId, KMccLinkGeneral, iLinkId, netSettings ) );
    
    iObserver.NetworkSetupStarted();
    }


// ---------------------------------------------------------------------------
// CMccNetworkSession::Matches
// ---------------------------------------------------------------------------
//
TBool CMccNetworkSession::Matches( TUint32 aSessionId, TUint32 aLinkId ) const
    {
    if ( aSessionId == iSessionId && aLinkId == iLinkId )
        {
        return ETrue;
        }
    else
        {
        return EFalse;
        }
    }


// ---------------------------------------------------------------------------
// CMccNetworkSession::HandleEvent
// ---------------------------------------------------------------------------
//
void CMccNetworkSession::HandleEvent( const TMccEvent& aEvent )
    {
    if ( KMccLinkCreated == aEvent.iEventType &&
         KErrNone == aEvent.iErrorCode )
        {
        iLinkUp = ETrue;
        iErrorCode = KErrNone;
        TInt count( iStreams.Count() );
        while( count )
            {
            count -= 1;
            TInt error = iStreams[count]->LinkReady();
            if ( KErrNone != error )
                {
                iErrorCode = error;
                break;
                }
            }
        
        iObserver.NetworkSetupReady( iErrorCode );
        }
    else if ( KMccLinkCreated == aEvent.iEventType &&
              KErrNone != aEvent.iErrorCode )
        {
        iLinkUp = EFalse;
        iErrorCode = aEvent.iErrorCode;
        iObserver.NetworkSetupReady( iErrorCode );
        }
    else if ( KMccEventCategoryStream == aEvent.iEventCategory &&
              KErrNone == aEvent.iErrorCode )
        {
        CStreamBase* strm( FindStream( aEvent.iStreamId ) );
        if ( strm )
            {
            strm->HandleEvent( aEvent );
            TControlState state = strm->State();
            if ( EStateChanging != state )
                {
                iObserver.StateChangeComplete( state, aEvent.iErrorCode );
                }
            }
        }
    else if ( KMccEventCategoryStream == aEvent.iEventCategory &&
              KErrNone != aEvent.iErrorCode )
        {
        // Do not forward the error event to streams as they are not ready
        // for handling those - try to tell the observer, user might be
        // able to help us.
        TControlState state = EStateChanging;
        CStreamBase* strm( FindStream( aEvent.iStreamId ) );
        if ( strm )
            {
            state = strm->State();
            }
        
        iObserver.ErrorDetected( state, aEvent.iErrorCode );
        }
    }


// ---------------------------------------------------------------------------
// CMccNetworkSession::SessionId
// ---------------------------------------------------------------------------
//
TUint32 CMccNetworkSession::SessionId() const
    {
    return iSessionId;
    }


// ---------------------------------------------------------------------------
// CMccNetworkSession::LinkId
// ---------------------------------------------------------------------------
//
TUint32 CMccNetworkSession::LinkId() const
    {
    return iLinkId;
    }


// ---------------------------------------------------------------------------
// CMccNetworkSession::AddStreamL
// ---------------------------------------------------------------------------
//
void CMccNetworkSession::AddStreamL( CStreamBase& aStream )
    {
    CStreamBase* stream = &aStream;
    
    // Assert that we have proper argument
    __ASSERT_ALWAYS( stream, User::Leave( KErrArgument ) );
    iStreams.AppendL( stream );
    }


// ---------------------------------------------------------------------------
// CMccNetworkSession::AddStreamL
// ---------------------------------------------------------------------------
//
void CMccNetworkSession::SetRemoteAddressL( const TInetAddr& aAddress )
    {
    // Assert the proper state before doing actual address setting
    __ASSERT_ALWAYS( ( iSessionId != KMaxTUint32 &&
            iLinkId != KMaxTUint32 && iLinkUp ),
            User::Leave( KErrNotReady ) );
    
    User::LeaveIfError(
            iMcc.SetRemoteAddress( iSessionId, iLinkId, aAddress ) );
    }


// ---------------------------------------------------------------------------
// CMccNetworkSession::FindStream
// ---------------------------------------------------------------------------
//
CStreamBase* CMccNetworkSession::FindStream( TUint32 aStreamId )
    {
    TInt idx( iStreams.Count() );
    while( idx )
        {
        idx -= 1;
        if ( iStreams[idx]->StreamId() == aStreamId )
            {
            return iStreams[idx];
            }
        }
    
    return NULL;
    }
